#include "mex.h"
#include "bspl.h"
#include <gsl/gsl_bspline.h>
#include <gsl/gsl_matrix.h>
#include <math.h>
#include <stdlib.h>
#include <float.h>
#include <string.h>

/**** global variables ****/

int mxIsVector(const mxArray *x)
{
	if(mxGetNumberOfDimensions(x) == 2 && (mxGetM(x) == 1 || mxGetN(x) == 1) && mxGetNumberOfElements(x) > 1)
		return 1;
	return 0;
}

void mexFunction(	int nlhs,	mxArray *plhs[],
						int nrhs,	const mxArray *prhs[])
{
	/***********
	 * B, B1, B2, B3 = bspl(knots, x)
	 * evaluate B-spline basis for N knots given by vector knots
	 * return the N B-spline functions and their first, second and third derivatives
	 * evaluated at locations x   
	 */

	const mwSize *mxN;
	const mxArray *mx_knots, *mx_x;
	mxArray *mx_B, *mx_B1, *mx_B2, *mx_B3;
	int n_knots, n_x;
	double *B, *B1, *B2, *B3;
	double *x, *knots;

	/***************
	 * check right number of arguments have been supplied
	 */
	if(nrhs < 2)
	{
		mexErrMsgTxt("bspl takes two arguments: B, B1, B2, B3 = bspl(knots, x)");
		return;
	}
	if(nlhs > 4)
	{
		mexErrMsgTxt("too many output arguments");
		return;
	}

	/***************
	 * check arguments have been supplied in the right format
	 */
	mx_knots = prhs[0];
	mx_x 		= prhs[1];
	
	/* knots must be a real vector of double values */
	if(!mxIsVector(mx_knots) || mxIsComplex(mx_knots) || !mxIsDouble(mx_knots))
	{
		mexErrMsgTxt("knots must be a real, double vector");
		return;
	}
	/* x must be a real vector of double values */
	if(!mxIsVector(mx_x) || mxIsComplex(mx_x) || !mxIsDouble(mx_x))
	{
		mexErrMsgTxt("knots must be a real, double vector");
		return;
	}
	
	/** allocate memory **/
	n_knots 		= mxGetNumberOfElements(mx_knots);
	n_x 			= mxGetNumberOfElements(mx_x);
	
	x 				= mxGetPr(mx_x);
	knots 		= mxGetPr(mx_knots);
	
	mx_B 			= mxCreateNumericMatrix(n_x, n_knots+2, mxDOUBLE_CLASS, mxREAL);
	mx_B1 		= mxCreateNumericMatrix(n_x, n_knots+2, mxDOUBLE_CLASS, mxREAL);
	mx_B2 		= mxCreateNumericMatrix(n_x, n_knots+2, mxDOUBLE_CLASS, mxREAL);
	mx_B3 		= mxCreateNumericMatrix(n_x, n_knots+2, mxDOUBLE_CLASS, mxREAL);
	B 				= mxGetPr(mx_B);
	B1 			= mxGetPr(mx_B1);
	B2 			= mxGetPr(mx_B2);
	B3 			= mxGetPr(mx_B3);	
	
	/* pass to BSPL routine */
	bspl(knots, x, n_knots, n_x, B, B1, B2, B3);
	
	/* save output */
	plhs[0] 		= mx_B;
	plhs[1] 		= mx_B1;
	plhs[2] 		= mx_B2;
	plhs[3] 		= mx_B3;	
	
	return;
}

void bspl(double *knots, double *x, int n_knots, int n_x, double *B, double *B1, double *B2, double *B3)
{
	gsl_bspline_workspace *ws;
	gsl_vector *bpts;
	gsl_matrix *Bi_deriv;
	int i, k, n_basis;
	double x_lb, x_ub;
	unsigned j, j_start, j_end;

	/* initialise bspline workspace
	 * cubic splines are specified by k=4 in the GSL B-spline library notation
	 * for n_knots internal break points, we will have n_knots+2 basis splines
	 */
	ws 		= gsl_bspline_alloc(4, n_knots); 
	n_basis 	= gsl_bspline_ncoeffs(ws); // = n_knots + 2
	
	/* define break points as internal knots */
	bpts 		= gsl_vector_alloc(n_knots);
	x_lb 		= knots[0];
	x_ub 		= knots[n_knots-1];
	for(k = 0; k < n_knots; ++k)
		gsl_vector_set(bpts, k, knots[k]);

	/* build */
	gsl_bspline_knots(bpts, ws);
	
	/* evaluate B-spline and its derivatives at each x*/
	Bi_deriv = gsl_matrix_alloc(4, 4); // k rows x 4 columns

	for(i = 0; i < n_x; ++i)
	{
		/* sanity check */
		if(x[i] < x_lb || x[i] > x_ub)
			continue;
		
		/* evalute non-zero B-spline derivatives 0-3 at x[i] */
		gsl_bspline_deriv_eval_nonzero(x[i], 3, Bi_deriv, &j_start, &j_end, ws);

		/* store non-zero spline and derivatives in array */
		for(j = j_start; j <= j_end; ++j)
		{
			B [SUB2IND_2D(i,j,n_x)] = gsl_matrix_get(Bi_deriv, j-j_start, 0);
			B1[SUB2IND_2D(i,j,n_x)] = gsl_matrix_get(Bi_deriv, j-j_start, 1);
			B2[SUB2IND_2D(i,j,n_x)] = gsl_matrix_get(Bi_deriv, j-j_start, 2);
			B3[SUB2IND_2D(i,j,n_x)] = gsl_matrix_get(Bi_deriv, j-j_start, 3);
		}
	}
	
	/* free bspline stuff */
	gsl_matrix_free(Bi_deriv);
	gsl_vector_free(bpts);
	gsl_bspline_free(ws);
}


